/*
 * imx-wm8960.c
 *
 * Copyright (C) 2012-2013 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/bitops.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/err.h>
#include <linux/irq.h>
#include <linux/io.h>
#include <linux/fsl_devices.h>
#include <linux/slab.h>
#include <linux/clk.h>
#include <linux/kthread.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <sound/jack.h>
#include <mach/dma.h>
#include <mach/clock.h>
#include <mach/audmux.h>
#include <mach/gpio.h>
#include <asm/mach-types.h>
#include <sound/wm8960.h>
#include "imx-ssi.h"
#include "../codecs/wm8960.h"

struct imx_priv {
	int sysclk;         /*mclk from the outside*/
	int codec_sysclk;
	int dai_hifi;
	int hp_gpio;
	int hp_status;
	int hp_active_low;
	int amic_gpio;
	int amic_status;
	struct platform_device *pdev;
};

unsigned int sample_format = SNDRV_PCM_FMTBIT_S16_LE;
static struct imx_priv card_priv;
static struct snd_soc_card snd_soc_card_imx;
static struct snd_soc_codec *gcodec;

static int imx_hifi_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct imx_priv *priv = &card_priv;
	struct mxc_audio_platform_data *plat = priv->pdev->dev.platform_data;

	if (!codec_dai->active)
		plat->clock_enable(1);

	return 0;
}

static void imx_hifi_shutdown(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct imx_priv *priv = &card_priv;
	struct mxc_audio_platform_data *plat = priv->pdev->dev.platform_data;

	if (!codec_dai->active)
		plat->clock_enable(0);

}

static const int bclk_divs[] = {
	1, -1, 2, 3, 4, -1, 6, 8, -1, 12, 16, -1, 24, 32, 32, 32
};

static struct {
	int sysclk;
	int rate;
	int div;
} adac_divs[] = {
	{ 12288000, 48000, 0 },
	{ 12288000, 32000, 1 },
	{ 12288000, 24000, 2 },
	{ 12288000, 16000, 3 },
	{ 12288000, 12000, 4 },
	{ 12288000, 8000,  6 },
	{ 11289600, 44100, 0 },
	{ 11289600, 22050, 2 },
	{ 11289600, 11025, 4 },
	{ 11289600, 8018,  5 },
	{ 2048000,  8000,  0 },
};

static int imx_hifi_hw_params(struct snd_pcm_substream *substream,
				     struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct imx_priv *priv = &card_priv;
	unsigned int channels = params_channels(params);
	unsigned int rate = params_rate(params);
	int bclk = snd_soc_params_to_bclk(params);
	int sysclk=11289600;
	int i,ret=0;	
	u32 dai_format;

	if (channels == 1)
		bclk *= 2;

	dai_format = SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBM_CFM ;

	/* set codec DAI configuration */
	ret = snd_soc_dai_set_fmt(codec_dai, dai_format);
	if (ret < 0){
		pr_err("-%s(): Cannot set codec_dai format \n", __FUNCTION__); 
 		return -EINVAL;
	}

	/* set i.MX active slot mask */
	snd_soc_dai_set_tdm_slot(cpu_dai,
				 channels == 1 ? 0xfffffffe : 0xfffffffc,
				 channels == 1 ? 0xfffffffe : 0xfffffffc,
				 2, 32);

	/* set cpu DAI configuration */
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0){
		pr_err("-%s(): Cannot set cpu_dai format \n", __FUNCTION__); 
		return -EINVAL;
	}


	/* find the sysclk rate && adc/dac div */
	for (i=0; i < ARRAY_SIZE(adac_divs); i++) {
		if (adac_divs[i].rate == rate) {
			sysclk = adac_divs[i].sysclk;
			snd_soc_dai_set_clkdiv(codec_dai, WM8960_ADCDIV, adac_divs[i].div);
			snd_soc_dai_set_clkdiv(codec_dai, WM8960_DACDIV, adac_divs[i].div);
			break;
		}
	}
	
	if (i == ARRAY_SIZE(adac_divs)) {
		pr_err("Rate %dHz is not support.\n",rate);
		return -EINVAL;
	}

	/* find the bclk div */
	for (i=0; i < ARRAY_SIZE(bclk_divs); i++) {
		if (bclk_divs[i] > 0) {
			if (sysclk/bclk_divs[i] == bclk) {
				snd_soc_dai_set_clkdiv(codec_dai, WM8960_BCLKDIV, i);
				break;
			}
		}
	}

	if (i == ARRAY_SIZE(bclk_divs)) {
		pr_err("Unsupported BCLK ratio %d\n", sysclk/bclk);
	}

	ret = snd_soc_dai_set_pll(codec_dai, 0, 0, priv->sysclk, sysclk); 
 	if( ret < 0 ){ 
 		pr_err("-%s(): Codec PLL setting error, %d\n", __FUNCTION__, ret); 
 		return ret; 
 	} 
 
 	return ret; 
}


static const struct snd_kcontrol_new controls[] = {
	SOC_DAPM_PIN_SWITCH("Headphone Jack"),
	SOC_DAPM_PIN_SWITCH("Ext Spk"),
	SOC_DAPM_PIN_SWITCH("My Mic"),
};

/* imx card dapm widgets */
static const struct snd_soc_dapm_widget imx_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headphone Jack", NULL),
	SND_SOC_DAPM_SPK("Ext Spk", NULL),
	SND_SOC_DAPM_MIC("My Mic", NULL),
};

/* imx machine connections to the codec pins */
static const struct snd_soc_dapm_route audio_map[] = {
	{ "Headphone Jack", NULL, "HP_L" },
	{ "Headphone Jack", NULL, "HP_R" },

	{ "Ext Spk", NULL, "SPK_RP" },
	{ "Ext Spk", NULL, "SPK_RN" },
	{ "Ext Spk", NULL, "SPK_LP" },
	{ "Ext Spk", NULL, "SPK_LN" },

	{ "LINPUT1", "MICB", "My Mic" },
};

static int imx_wm8960_init(struct snd_soc_pcm_runtime *rtd)
{
	struct snd_soc_codec *codec = rtd->codec;
	//struct imx_priv *priv = &card_priv;
	//struct platform_device *pdev = priv->pdev;
	//struct mxc_audio_platform_data *plat = pdev->dev.platform_data;
	u16 reg;
	int ret = 0;

	gcodec = rtd->codec;

	/* Add imx specific widgets */
	snd_soc_dapm_new_controls(&codec->dapm, imx_dapm_widgets,
				  ARRAY_SIZE(imx_dapm_widgets));

	ret = snd_soc_add_controls(codec, controls, ARRAY_SIZE(controls));
	if (ret<0)
		pr_err("Failed to set codec controls: %d\n", ret);	

	/* Set up imx specific audio path audio_map */
	snd_soc_dapm_add_routes(&codec->dapm, audio_map, ARRAY_SIZE(audio_map));

	snd_soc_dapm_enable_pin(&codec->dapm, "Headphone Jack");
	snd_soc_dapm_enable_pin(&codec->dapm, "Ext Spk");
	snd_soc_dapm_enable_pin(&codec->dapm, "My Mic");

	snd_soc_dapm_sync(&codec->dapm);

	//set ADCLRC pin to gpio fuction
	reg = snd_soc_read(codec, WM8960_IFACE2);
	snd_soc_write(codec, WM8960_IFACE2, reg | 0x40);

	//set LRCM =1
	reg = snd_soc_read(codec, WM8960_ADDCTL2);
	snd_soc_write(codec, WM8960_ADDCTL2, reg | 0x04);


	return 0;
}

static struct snd_soc_ops imx_hifi_ops = {
	.startup = imx_hifi_startup,
	.shutdown = imx_hifi_shutdown,
	.hw_params = imx_hifi_hw_params,
};

static struct snd_soc_dai_link imx_dai[] = {
	{
		.name = "HiFi",
		.stream_name = "HiFi",
		.codec_dai_name	= "wm8960",
		.codec_name	= "wm8960.0-001a",
		.cpu_dai_name	= "imx-ssi.1",
		.platform_name	= "imx-pcm-audio.1",
		.init		= imx_wm8960_init,
		.ops		= &imx_hifi_ops,
	},
};

static struct snd_soc_card snd_soc_card_imx = {
	.name		= "wm8960-audio",
	.dai_link	= imx_dai,
	.num_links	= ARRAY_SIZE(imx_dai),
};

static int imx_audmux_config(int slave, int master)
{
	unsigned int ptcr, pdcr;
	slave = slave - 1;
	master = master - 1;

	ptcr = MXC_AUDMUX_V2_PTCR_SYN |
		MXC_AUDMUX_V2_PTCR_TFSDIR |
		MXC_AUDMUX_V2_PTCR_TFSEL(master) |
		MXC_AUDMUX_V2_PTCR_TCLKDIR |
		MXC_AUDMUX_V2_PTCR_TCSEL(master);
	pdcr = MXC_AUDMUX_V2_PDCR_RXDSEL(master);
	mxc_audmux_v2_configure_port(slave, ptcr, pdcr);

	ptcr = MXC_AUDMUX_V2_PTCR_SYN;
	pdcr = MXC_AUDMUX_V2_PDCR_RXDSEL(slave);
	mxc_audmux_v2_configure_port(master, ptcr, pdcr);

	return 0;
}

/*
 * This function will register the snd_soc_pcm_link drivers.
 */
static int __devinit imx_wm8960_probe(struct platform_device *pdev)
{

	struct mxc_audio_platform_data *plat = pdev->dev.platform_data;
	struct imx_priv *priv = &card_priv;
	int ret = 0;

	priv->pdev = pdev;

	imx_audmux_config(plat->src_port, plat->ext_port);

	if (plat->init && plat->init()) {
		ret = -EINVAL;
		return ret;
	}

	

	priv->sysclk = plat->sysclk;
	priv->hp_gpio = plat->hp_gpio;
	priv->hp_active_low = plat->hp_active_low;
	return ret;
}

static int __devexit imx_wm8960_remove(struct platform_device *pdev)
{
	struct mxc_audio_platform_data *plat = pdev->dev.platform_data;
	//struct imx_priv *priv = &card_priv;

	if (plat->finit)
		plat->finit();

	return 0;
}

static struct platform_driver imx_wm8960_driver = {
	.probe = imx_wm8960_probe,
	.remove = imx_wm8960_remove,
	.driver = {
		   .name = "imx-wm8960",
		   .owner = THIS_MODULE,
		   },
};

static struct platform_device *imx_snd_device;

static int __init imx_asoc_init(void)
{
	int ret;

	ret = platform_driver_register(&imx_wm8960_driver);
	if (ret < 0)
		goto exit;

	imx_dai[0].codec_name = "wm8960.0-001a";

	imx_snd_device = platform_device_alloc("soc-audio", 5);
	if (!imx_snd_device)
		goto err_device_alloc;

	platform_set_drvdata(imx_snd_device, &snd_soc_card_imx);

	ret = platform_device_add(imx_snd_device);

	if (0 == ret)
		goto exit;

	platform_device_put(imx_snd_device);

err_device_alloc:
	platform_driver_unregister(&imx_wm8960_driver);
exit:
	return ret;
}

static void __exit imx_asoc_exit(void)
{
	platform_driver_unregister(&imx_wm8960_driver);
	platform_device_unregister(imx_snd_device);
}

module_init(imx_asoc_init);
module_exit(imx_asoc_exit);

/* Module information */
MODULE_DESCRIPTION("ALSA SoC imx wm8960");
MODULE_LICENSE("GPL");
